open Syntax

(* environnements *)
type 'a env = symbol -> 'a
let empty_env: 'a env = fun _ -> failwith "Unbound_Variable"
let get_env (e: 'a env) (x: symbol): 'a =  e x
let add_env (e: 'a env) (x: symbol) (v: 'a): 'a env =
  fun y -> if x=y then v else e y

(* mémoire *)
type address = int
type 'a mem = 'a list
let empty_mem: 'a mem = []
let get_mem (m: 'a mem) (r: address): 'a = 
  try List.nth m r with _ -> failwith "Invalid Memory Access"
let rec set_mem (m: 'a mem) (r: address) (v: 'a): 'a mem = 
  match m, r with
    | [],_ -> failwith "Invalid Memory Access"
    | _::m, 0 -> v::m
    | w::m, r -> w::set_mem m (r-1) v
let rec alloc_mem (m: 'a mem) (v: 'a): address * 'a mem =
  match m with
    | [] -> 0, [v]
    | x::q -> let r,m = alloc_mem q v in r+1,x::m

(* valeurs *)
type value = 
  | VUnit
  | VInt of int
  | VFun of symbol * value env * expr
  | VRef of address


(* évaluation *)
open Syntax
let rec eval env mem = function
  | TUnit -> VUnit, mem
  | TInt i -> VInt i, mem
  | TVar x -> get_env env x, mem
  | TFun(x,b) -> VFun(x,env,b), mem
  | TSeq(a,b) -> 			(* a;b *)
      let a,mem = eval env mem a in
      let b,mem = eval env mem b in
	b,mem
  | TAdd(a,b) -> 			(* a+b *)
      let a,mem = eval_int env mem a in
      let b,mem = eval_int env mem b in
	VInt(a+b),mem
  | TMul(a,b) -> 			(* a*b *)
      let a,mem = eval_int env mem a in
      let b,mem = eval_int env mem b in
	VInt(a+b),mem
  | TLet(x,a,b) -> 			(* let x=a in b *)
      let a,mem = eval env mem a in
	eval (add_env env x a) mem b 

  | TApp(a,b) -> 			(* (a b) *)
      let a,mem = eval env mem a in
      let b,mem = eval env mem b in
	(match a with
	  | VFun(x,env,k) -> eval (add_env env x b) mem k
	  | _ -> failwith "Function expected")
  | TRef a -> 				(* ref a *)
      let a,mem = eval env mem a in
      let r,mem = alloc_mem mem a in
	VRef r, mem
  | TDeref r -> 			(* !r *)
      let r,mem = eval_ref env mem r in
	get_mem mem r, mem
  | TAffect(r,b) -> 			(* r:=b *)
      let r,mem = eval_ref env mem r in
      let b,mem = eval env mem b in
      let mem = set_mem mem r b in
	VUnit, mem
and eval_int env mem a = 
  match eval env mem a with
    | VInt i,mem -> i,mem
    | _ -> failwith "Int value expected"
and eval_ref env mem a = 
  match eval env mem a with
    | VRef r,mem -> r,mem
    | _ -> failwith "Ref value expected"

(* affichage *)
let rec print_value f (v,m) = match v with 
  | VUnit -> Format.fprintf f "()"
  | VInt i -> Format.fprintf f "%i" i
  | VRef a -> Format.fprintf f "ref (%a)" print_value (get_mem m a,m)
  | VFun _ -> Format.fprintf f "<fun>"
